# TPK4186 - Advanced Tools for Performance Engineering
# Calculator.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Vectors
# 3. Calculator
#  3.1. Creator
#  3.2. Preprocessing of the model
#  3.3. Choice of dt
#  3.4. Operations on Vectors
#  3.5. Matrix Exponentiation
#  3.6. Assessment

# 1. Required modules
# -------------------

import sys
import re
import math
from MarkovModels import *
from TextInterface import *

# 2. Vectors
# ----------

class Vector:
  def __init__(self, size):
    self.size = size
    self.values = [0.0 for _ in range(0, self.size)]

  def GetValue(self, index):
    if index<0 or index>=self.size:
      return None
    return self.values[index]

  def SetValue(self, index, value):
    if index<0 or index>=self.size:
      return
    self.values[index] = value

  def AddValue(self, index, value):
    if index<0 or index>=self.size:
      return
    self.values[index] += value

  def RemoveValue(self, index, value):
    if index<0 or index>=self.size:
      return
    self.values[index] -= value

  def FillWithScalar(self, k):
    for i in range(0, self.size):
      self.values[i] = k

  def DivideByScalar(self, k):
    for i in range(0, self.size):
      self.values[i] = self.values[i] / k

  def SetWithVector(self, V):
    for i in range(0, self.size):
      self.values[i] = V.values[i]

  def AddVector(self, V):
    for i in range(0, self.size):
      self.values[i] += V.values[i]

  def SetWithSumOfVectors(self, V, W):
    for i in range(0, self.size):
      self.values[i] = V.values[i] + W.values[i]

  def IsNearlyEqualToVector(self, V, epsilon):
    maxDiff = 0
    for i in range(0, self.size):
      u = self.values[i]
      v = V.values[i]
      if v==0:
        if u==0:
          diff = 0
        else:
          diff = u
      else:
        diff = math.fabs(v-u)/v
      if diff>maxDiff:
        maxDiff = diff
    return maxDiff <= epsilon

  def AddTemporizedMean(self, P, Q, dt):
    for i in range(0, self.size):
      self.values[i] += (P.values[i]+Q.values[i])*dt/2

  def Dump(self, file):
    file.write("[")
    for index in range(0, self.size):
      if index>0:
        file.write(", ")
      file.write("{0:g}".format(self.values[index]))
    file.write("]\n")
    file.flush()

# 3. Calculator
# -------------

# 3.1. Creator
# ------------

class Calculator:
  def __init__(self):
    self.model = None
    self.numberOfStates = 0
    self.error = 0
    self.rho = 0.1
    self.epsilon = 1.0e-9
    self.minimumNumberOfPointsInPeriod = 10
    self.probabilities = None
    self.sojournTimes = None
    self.outputFile = None

# 3.2. Preprocessing of the model
# -------------------------------

  def PreprocessModel(self):
    self.error = 0
    self.CalculateValueOfParameters()
    self.CalculateValueOTransitionRates()
    self.CalculateValueOfRewardExpressions()
    self.SetIndicesOfStates()
    return self.error

  def CalculateValueOfParameters(self):
    for parameter in self.model.GetParameters():
      parameter.SetValue(None)
    for parameter in self.model.GetParameters():
      self.CalculateValueOfParameter(parameter)

  def CalculateValueOfParameter(self, parameter):
    value = parameter.GetValue()
    if value==None:
      value = self.CalculateValueOfExpression(parameter.GetDefinition())
      parameter.SetValue(value)
    return value

  def CalculateValueOTransitionRates(self):
    for transition in self.model.GetTransitions():
      rateExpression = transition.GetRateExpression()
      self.CalculateValueOfExpression(rateExpression)

  def CalculateValueOfRewardExpressions(self):
    for reward in self.model.GetRewards():
      for state in reward.GetStates():
        expression = reward.GetValueOfState(state)
        self.CalculateValueOfExpression(expression)

  def CalculateValueOfExpression(self, expression):
    if expression.GetType()==Expression.PARAMETER:
      parameter = expression.GetParameter()
      value = self.CalculateValueOfParameter(parameter)
      expression.SetValue(value)
    elif expression.GetType()==Expression.REAL:
      value = expression.GetValue()
    elif expression.GetType()==Expression.ADD:
      value = 0
      for child in expression.GetChildren():
        childValue = self.CalculateValueOfExpression(child)
        value += childValue
      expression.SetValue(value)
    elif expression.GetType()==Expression.SUB:
      value = 0
      first = True
      for child in expression.GetChildren():
        childValue = self.CalculateValueOfExpression(child)
        if first:
          value = childValue
          first = False
        else:
          value -= childValue
      expression.SetValue(value)
    elif expression.GetType()==Expression.MUL:
      value = 1
      for child in expression.GetChildren():
        childValue = self.CalculateValueOfExpression(child)
        value *= childValue
      expression.SetValue(value)
    elif expression.GetType()==Expression.DIV:
      value = 1
      first = True
      for child in expression.GetChildren():
        childValue = self.CalculateValueOfExpression(child)
        if first:
          value = childValue
          first = False
        else:
          value /= childValue
      expression.SetValue(value)
    else: # Cannot happen
      value = None
    return value

  def SetIndicesOfStates(self):
    index = 0
    for state in self.model.GetStates():
      state.SetIndex(index)
      index += 1

# 3.3. Choice of dt
# -----------------

  def GetRho(self):
    return self.rho

  def SetRho(self, rho):
    self.rho = rho

  def GetMinimumNumberOfPointsInPeriod(self):
    return self.minimumNumberOfPointsInPeriod

  def SetMinimumNumberOfPointsInPeriod(self, numberOfPoints):
    self.minimumNumberOfPointsInPeriod = numberOfPoints

  def SelectDt(self, startTime, endTime):
    Qiis = [0 for _ in range(0, self.model.GetNumberOfStates())]
    for transition in self.model.GetTransitions():
      sourceState = transition.GetSourceState()
      sourceStateIndex = sourceState.GetIndex()
      rateExpression = transition.GetRateExpression()
      value = rateExpression.GetValue()
      Qiis[sourceStateIndex] += value
    maxQii = 0
    for i in range(0, self.model.GetNumberOfStates()):
      if Qiis[i]>maxQii:
        maxQii = Qiis[i]
    dt1 = self.rho/maxQii
    period = endTime - startTime
    dt2 = period/self.minimumNumberOfPointsInPeriod
    if dt1>=1 and dt2>=1:
      return 1
    if dt2<dt1:
      return dt2
    k = math.ceil(period/dt1)
    return period/k

  def CalculateProbabilitiesOfTransitions(self, dt):
    for transition in self.model.GetTransitions():
      rateExpression = transition.GetRateExpression()
      rate = rateExpression.GetValue()
      transition.SetProbability(dt*rate)

# 3.4. Operations on Vectors
# --------------------------

  def GetEpsilon(self):
    return self.epsilon

  def SetEpsilon(self, epsilon):
    self.epsilon = epsilon

  def NewProbabilityVector(self):
    return Vector(self.model.GetNumberOfStates())

  def FillProbabilityVectorWithInitialProbabilities(self, probabilities):
    for state in self.model.GetStates():
      probabilities.SetValue(state.GetIndex(), state.GetInitialProbability())

# 3.5. Matrix Exponentiation
# --------------------------

  def MatrixExponentiation(self, startTime, endTime):
    dt = self.SelectDt(startTime, endTime)
    self.CalculateProbabilitiesOfTransitions(dt)
    P1 = self.NewProbabilityVector()
    P0 = self.NewProbabilityVector()
    K1 = self.NewProbabilityVector()
    K0 = self.NewProbabilityVector()
    P0.SetWithVector(self.probabilities)
    K0.SetWithVector(self.probabilities)
    # probabilities.Dump(sys.stdout)
    convergence = False
    time = startTime
    while not convergence and time<endTime:
      time += dt
      # print("time {0:g}".format(time))
      self.UpdateCurrentProbabilitiesWithSeriesDevelopment(P1, P0, K1, K0)
      self.sojournTimes.AddTemporizedMean(self.probabilities, P1, dt)
      convergence = self.probabilities.IsNearlyEqualToVector(P1, self.epsilon)
      self.probabilities.SetWithVector(P1)
      # probabilities.Dump(sys.stdout)
      if convergence:
        break
      P0.SetWithVector(self.probabilities)
      K0.SetWithVector(self.probabilities)
    if time<endTime:
      self.sojournTimes.AddTemporizedMean(self.probabilities, P1, endTime-time)

  def UpdateCurrentProbabilitiesWithSeriesDevelopment(self, P1, P0, K1, K0):
    k = 0
    converged = False
    while not converged:
      k += 1
      # print("k {0:d}".format(k))
      K1.FillWithScalar(0.0)
      self.MultiplyRightProbabilityMatrixByColumnVector(K1, K0)
      K1.DivideByScalar(k)
      P1.SetWithSumOfVectors(P0, K1)
      converged = P1.IsNearlyEqualToVector(P0, self.epsilon)
      P0.SetWithVector(P1)
      K0.SetWithVector(K1)

  def MultiplyRightProbabilityMatrixByColumnVector(self, R, V):
    for transition in self.model.GetTransitions():
      p = transition.GetProbability()
      i = transition.GetSourceState().GetIndex()
      j = transition.GetTargetState().GetIndex()
      dp = p*V.GetValue(i)
      R.RemoveValue(i, dp)
      R.AddValue(j, dp)

# 3.6. Assessment
# ---------------

  def AssessModelAndExportResults(self, model, missionTimes, fileName):
    try:
      file = open(fileName, "w")
    except:
      sys.stderr.write("Unable to open file {0:s}\n".format(fileName))
      sys.stderr.flush()
      sys.exit()
    self.AssessModelAndPrintResults(model, missionTimes, file)
    file.flush()
    file.close()

  def AssessModelAndPrintResults(self, model, missionTimes, file):
    missionTimes.sort()
    if len(missionTimes)==0:
      return
    self.model = model
    self.numberOfStates = self.model.GetNumberOfStates()
    self.PreprocessModel()
    self.probabilities = self.NewProbabilityVector()
    self.sojournTimes = self.NewProbabilityVector()
    self.outputFile = file
    startTime = 0.0
    self.FillProbabilityVectorWithInitialProbabilities(self.probabilities)
    # self.probabilities.Dump(sys.stdout)
    self.sojournTimes.FillWithScalar(0.0)
    for endTime in missionTimes:
      if endTime==0:
        continue
      self.AssessModelInTimePeriodAndPrintResults(startTime, endTime, file)
      startTime = endTime

  def AssessModelInTimePeriodAndPrintResults(self, startTime, endTime, file):
    self.MatrixExponentiation(startTime, endTime)
    file.write("period\t{0:g}\t{1:g}\n".format(startTime, endTime))
    file.write("state\tprobability\tsojourn-time\n")
    for state in self.model.GetStates():
      name = state.GetName()
      index = state.GetIndex()
      file.write("{0:s}\t{1:g}\t{2:g}\n".format(name, self.probabilities.GetValue(index), self.sojournTimes.GetValue(index)))
    if len(self.model.GetRewards())!=0:
      file.write("reward\tvalue\taverage-value\ttotal\n")
    for reward in self.model.GetRewards():
      name = reward.GetName()
      value = 0
      sumValues = 0
      for state in reward.GetStates():
        stateValue = reward.GetValueOfState(state).GetValue()
        stateIndex = state.GetIndex()
        value += self.probabilities.GetValue(stateIndex)*stateValue
        sumValues += self.sojournTimes.GetValue(stateIndex)*stateValue
      file.write("{0:s}\t{1:g}\t{2:g}\t{3:g}\n".format(name, value, sumValues/endTime, sumValues))
    file.write("\n")
