# TPK4186 - Advanced Tools for Performance Engineering
# MarkovModels.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Expressions
# 3. Parameters
# 4. States
# 5. Transitions
# 6. Rewards
# 7. Markov Models

# 1. Required modules
# -------------------

import sys
import re

# 2. Expressions
# -------------

class Expression:
  PARAMETER = 1
  REAL = 2
  ADD = 3
  SUB = 4
  MUL = 5
  DIV = 6

  def __init__(self, type, children = []):
    self.type = type
    self.parameterName = None
    self.parameter = None
    self.children = children[:]
    self.value = None

  @staticmethod
  def NewParameterReference(parameterName):
    self = Expression(Expression.PARAMETER)
    self.parameterName = parameterName
    return self

  @staticmethod
  def NewReal(value):
    self = Expression(Expression.REAL)
    self.value = value
    return self

  @staticmethod
  def NewAdd(children = []):
    self = Expression(Expression.ADD, children)
    return self

  @staticmethod
  def NewSub(children = []):
    self = Expression(Expression.SUB, children)
    return self

  @staticmethod
  def NewMul(children = []):
    self = Expression(Expression.MUL, children)
    return self

  @staticmethod
  def NewDiv(children = []):
    self = Expression(Expression.DIV, children)
    return self

  def GetType(self):
    return self.type

  def GetParameterName(self):
    return self.parameterName

  def SetParameterName(self, name):
    self.parameterName = name

  def GetParameter(self):
    return self.parameter

  def SetParameter(self, parameter):
    self.parameter = parameter

  def GetValue(self):
    return self.value

  def SetValue(self, value):
    self.value = value

  def GetChildren(self):
    return self.children

  def AddChild(self, child):
    self.children.append(child)

# 3. Parameters
# -------------

class Parameter:
  def __init__(self, name):
    self.name = name
    self.definition = 0
    self.value = None

  def GetName(self):
    return self.name

  def GetDefinition(self):
    return self.definition

  def SetDefinition(self, definition):
    self.definition = definition

  def GetValue(self):
    return self.value

  def SetValue(self, value):
    self.value = value

# 4. States
# ---------

class State:
  def __init__(self, name):
    self.name = name
    self.initialProbability = 0.0
    self.index = -1

  def GetName(self):
    return self.name

  def GetInitialProbability(self):
    return self.initialProbability

  def SetInitialProbability(self, value):
    self.initialProbability = value

  def GetIndex(self):
    return self.index

  def SetIndex(self, index):
    self.index = index

# 5. Transitions
# --------------

class Transition:
  def __init__(self, sourceState, targetState, rateExpression):
    self.sourceState = sourceState
    self.targetState = targetState
    self.rateExpression = rateExpression
    self.probability = 0

  def GetSourceState(self):
    return self.sourceState

  def GetTargetState(self):
    return self.targetState

  def GetRateExpression(self):
    return self.rateExpression

  def GetProbability(self):
    return self.probability

  def SetProbability(self, probability):
    self.probability = probability


# 6. Rewards
# ----------

class Reward:
  def __init__(self, name):
    self.name = name
    self.states = dict()

  def GetName(self):
    return self.name

  def GetStates(self):
    return self.states

  def GetValueOfState(self, state):
    return self.states.get(state, None)

  def SetValueOfState(self, state, value):
    self.states[state] = value

# 7. Markov Models
# ----------------

class MarkovModel:
  def __init__(self, name):
    self.name = name
    self.parameters = dict()
    self.states = dict()
    self.transitions = []
    self.rewards = dict()

  def GetName(self):
    return self.name

  def GetParameters(self):
    return self.parameters.values()

  def LookForParameter(self, name):
    return self.parameters.get(name, None)

  def NewParameter(self, name, definition = None):
    parameter = Parameter(name)
    parameter.SetDefinition(definition)
    self.parameters[name] = parameter
    return parameter

  def GetStates(self):
    return self.states.values()

  def GetNumberOfStates(self):
    return len(self.states)

  def LookForState(self, name):
    return self.states.get(name, None)

  def NewState(self, name):
    state = State(name)
    self.states[name] = state
    return state

  def GetTransitions(self):
    return self.transitions

  def NewTransition(self, sourceState, targetState, rateExpression):
    transition = Transition(sourceState, targetState, rateExpression)
    self.transitions.append(transition)
    return transition

  def GetRewards(self):
    return self.rewards.values()

  def LookForReward(self, name):
    return self.rewards.get(name, None)

  def NewReward(self, name):
    reward = Reward(name)
    self.rewards[name] = reward
    return reward
