# TPK4186 - Advanced Tools for Performance Engineering
# generator.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

import sys
import math

def ExportModel(fileName):
  try:
    file = open(fileName, "w")
  except:
    sys.stderr.write("Unable to open file {0:s}\n".format(fileName))
    sys.stderr.flush()
    sys.exit()
  PrintModel(file)
  file.flush()
  file.close()

def PrintModel(file):
  file.write("model OilAndGasFacility\n")
  DeclareStates(file)
  DeclareTransitions(file)
  DeclareParameters(file)
  DeclareReward(file)
  file.write("end\n")

def DeclareStates(file):
  # states are composed of four elements:
  # - number of HPS failed in {0, 1, 2, 3}
  # - number of DEH failed in {0, 1, 2}
  # - number of CMP failed in {0, 1, 2}
  # - state of the MUP in {0, 1} (1:working, 2:failed}
  for HPS in range(0, 4):
    for DEH in range(0, 3):
      for CMP in range(0, 3):
        for MUP in range(0, 2):
          DeclareState(HPS, DEH, CMP, MUP, file)

def DeclareState(HPS, DEH, CMP, MUP, file):
  file.write("\tstate {0:s}".format(GetStateName(HPS, DEH, CMP, MUP)))
  if HPS==0 and DEH==0 and CMP==0 and MUP==0:
    file.write(" (initial = 1.0)")
  file.write(";\n")

def GetStateName(HPS, DEH, CMP, MUP):
  return "s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP)

def DeclareParameters(file):
  file.write("\tparameter lambdaHPS = 8.91e-5;\n")
  file.write("\tparameter muHPS = 2.54e-3;\n")
  file.write("\tparameter lambdaDEH = 3.11e-5;\n")
  file.write("\tparameter muDEH = 3.95e-3;\n")
  file.write("\tparameter lambdaCMP = 3.50e-5;\n")
  file.write("\tparameter muCMP = 5.14e-3;\n")
  file.write("\tparameter lambdaMUP = 1.00e-2;\n")

def DeclareTransitions(file):
  # states are composed of four elements:
  # - number of HPS failed in {0, 1, 2, 3}
  # - number of DEH failed in {0, 1, 2}
  # - number of CMP failed in {0, 1, 2}
  # - state of the MUP in {0, 1, 2} (0: standby, 1:working, 2:failed}
  for HPS in range(0, 4):
    for DEH in range(0, 3):
      for CMP in range(0, 3):
        for MUP in range(0, 2):
          if HPS!=3:
            DeclareFailureHPS(HPS, DEH, CMP, MUP, file)
          if HPS!=0:
            DeclareRepairHPS(HPS, DEH, CMP, MUP, file)
          if DEH!=2:
            DeclareFailureDEH(HPS, DEH, CMP, MUP, file)
          if DEH!=0:
            DeclareRepairDEH(HPS, DEH, CMP, MUP, file)
          if CMP!=2:
            DeclareFailureCMP(HPS, DEH, CMP, MUP, file)
          if CMP!=0:
            DeclareRepairCMP(HPS, DEH, CMP, MUP, file)
          if (HPS==3 or DEH==2) and MUP==0:
            DeclareFailureMUP(HPS, DEH, CMP, MUP, file)

def DeclareFailureHPS(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS+1, DEH, CMP, MUP))
  file.write(" (* lambdaHPS {0:d})".format(3-HPS))
  file.write(";\n")

def DeclareRepairHPS(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS-1, DEH, CMP, MUP))
  file.write(" (* muHPS {0:d})".format(HPS))
  file.write(";\n")

def DeclareFailureDEH(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH+1, CMP, MUP))
  file.write(" (* lambdaDEH {0:d})".format(2-DEH))
  file.write(";\n")

def DeclareRepairDEH(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH-1, CMP, MUP))
  file.write(" (* muDEH {0:d})".format(DEH))
  file.write(";\n")

def DeclareFailureCMP(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP+1, MUP))
  file.write(" (* lambdaCMP {0:d})".format(2-DEH))
  file.write(";\n")

def DeclareRepairCMP(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP, MUP))
  file.write(" s{0:d}{1:d}{2:d}{3:d}".format(HPS, DEH, CMP-1, MUP))
  file.write(" (* muCMP {0:d})".format(CMP))
  file.write(";\n")

def DeclareFailureMUP(HPS, DEH, CMP, MUP, file):
  file.write("\ttransition")
  file.write(" s{0:d}{1:d}{2:d}0".format(HPS, DEH, CMP))
  file.write(" s{0:d}{1:d}{2:d}1".format(HPS, DEH, CMP))
  file.write(" lambdaMUP")
  file.write(";\n")

def DeclareReward(file):
  # states are composed of four elements:
  # - number of HPS failed in {0, 1, 2, 3}
  # - number of DEH failed in {0, 1, 2}
  # - number of CMP failed in {0, 1, 2}
  # - state of the MUP in {0, 1} (1:working, 2:failed}
  first = True
  file.write("\treward production =")
  for HPS in range(0, 4):
    for DEH in range(0, 3):
      for CMP in range(0, 3):
        for MUP in range(0, 2):
          capacityHPS = (3-HPS)*55
          capacityDEH = (2-DEH)*65
          capacityCMP = (2-CMP)*52
          capacityMUP = (1-MUP)*100
          capacityHPSDEH = min([capacityHPS, capacityDEH])
          if capacityHPSDEH==0:
            capacityHPSDEHMUP = capacityMUP
          else:
            capacityHPSDEHMUP = capacityHPSDEH
          capacity = min([capacityHPSDEHMUP, capacityCMP])
          production = min([capacity, 100])
          if first:
            first = False
          else:
            file.write(",")
          file.write("\n\t\t{0:s}: {1:d}".format(GetStateName(HPS, DEH, CMP, MUP), production))
  file.write(";\n")

ExportModel("OilAndGasFacility.mrk")